package experimentalcode.tobias.TrigonometricPruning;

import java.util.ArrayList;
import java.util.List;
import java.util.PriorityQueue;
import java.util.TreeMap;

import de.lmu.ifi.dbs.elki.data.DoubleVector;
import de.lmu.ifi.dbs.elki.data.HyperBoundingBox;
import de.lmu.ifi.dbs.elki.database.ids.DBID;
import de.lmu.ifi.dbs.elki.index.tree.spatial.SpatialEntry;
import de.lmu.ifi.dbs.elki.index.tree.spatial.SpatialPointLeafEntry;
import de.lmu.ifi.dbs.elki.index.tree.spatial.rstarvariants.rstar.RStarTree;
import de.lmu.ifi.dbs.elki.index.tree.spatial.rstarvariants.rstar.RStarTreeNode;
import experimentalcode.tobias.TrigonometricPruning.structures.DataBall;
import experimentalcode.tobias._index.helpstructures.KNNList;
import experimentalcode.tobias.rknnjoin.DistanceCalc;
import experimentalcode.tobias.rknnjoin.structures.ObjectScorePair;
import experimentalcode.tobias.rknnjoin.structures.buffer.Buffer2;


//Join according to zhang, mamoulis, papadias, tao paper on an RTree
public class BNN extends AkNNJoin{
  
  
  RStarTree<DoubleVector> rTree;

  Buffer2<SpatialEntry> lruBuffer;
  
  public BNN(RStarTree<DoubleVector> rTree, int bufferSize){
    this.rTree = rTree;
    lruBuffer = new Buffer2<SpatialEntry>(bufferSize);
  }

  @Override
  public TreeMap<DBID,KNNList<DoubleVector>> join(DataBall queryBall, int k){
   
    TreeMap<DBID,KNNList<DoubleVector>> aknnJoinResult = new TreeMap<DBID,KNNList<DoubleVector>>();
    
    HyperBoundingBox queryBox = queryBall.getMBR();
    //Organize pages by mindist
    PriorityQueue<ObjectScorePair<SpatialEntry>> apl = new PriorityQueue<ObjectScorePair<SpatialEntry>>();
    apl.add(new ObjectScorePair(rTree.getRootEntry(), DistanceCalc.mindist(queryBox, rTree.getRootEntry())));
    distanceCalculations++;

    double globalKnnDist = Double.MAX_VALUE;
    while(!apl.isEmpty()){
      double minDist = apl.peek().score;
      if(globalKnnDist < minDist)break;
      
      SpatialEntry entry = apl.poll().object;
      RStarTreeNode node = rTree.getNode(entry);
      
      logicalPageAccesses++;
      if(lruBuffer.add(entry))
        physicalPageAccesses++;
      
      
      if(!node.getEntries().get(0).isLeafEntry()){ //Directory Node
        for(SpatialEntry se:node.getEntries()){
          if(se == null) continue;
          double dist = DistanceCalc.mindist(queryBox, se);
          distanceCalculations++;
          apl.add(new ObjectScorePair(se, dist));
        }
      }else{ //Data Node
        /*for(DoubleVector dv:queryBall.objects){
          KNNList knnlist = aknnJoinResult.get(dv.getID());
          
          //If node can not contain better results continue
          if(knnlist != null){
            distanceCalculations++;
            if(knnlist.getKNNdist() < mindist(dv, entry))
                continue;
          }
          for(SpatialEntry se2:node.getEntries()){
            if(se2 == null) continue;
            SpatialLeafEntry sle = (SpatialLeafEntry)se2;
            distanceCalculations++;
            if(DistanceCalc.mindist(sle.getValues(), queryBox)>globalKnnDist)
              continue;
            
            DoubleVector t = new DoubleVector(sle.getValues());
            t.setID(sle.getID());
            addResult(dv, t, k, aknnJoinResult);
          }
          //update global KNNDist
          globalKnnDist = 0.0;
          for(DoubleVector temp:queryBall.objects){
            KNNList knnlist2 = aknnJoinResult.get(temp.getID());
            if(knnlist2 != null){
              globalKnnDist = Math.max(knnlist2.getKNNdist(), globalKnnDist);
            }else{
              globalKnnDist = Double.MAX_VALUE;
              break;
            }
          }
        }*/
        //Remove all query vectors which can not be updated
        List<DoubleVector> updateQuery = new ArrayList<DoubleVector>();
        for(DoubleVector dv:queryBall.objects){
          KNNList knnlist = aknnJoinResult.get(dv.getID());
          if(knnlist != null){
            distanceCalculations++;
            if(knnlist.getKNNdist() < mindist(dv, entry))
                continue; 
          }
          updateQuery.add(dv);
        }
        //Remove all data vectors which can not contribute new knns
        List<DoubleVector> kNNcandidates = new ArrayList<DoubleVector>();
        for(SpatialEntry se2:node.getEntries()){
          if(se2 == null) continue;
          SpatialPointLeafEntry sle = (SpatialPointLeafEntry)se2;
          distanceCalculations++;
          if(DistanceCalc.mindist(sle.getValues(), queryBox)>globalKnnDist)
            continue;
          else{
            DoubleVector t = new DoubleVector(sle.getValues());
            t.setID(sle.getDBID());
            kNNcandidates.add(t);
          }
        }
        
        for(DoubleVector dv:updateQuery){
          for(DoubleVector dvs:kNNcandidates){
            addResult(dv, dvs, k, aknnJoinResult);
          }
        }
      //update global KNNDist
        globalKnnDist = 0.0;
        for(DoubleVector temp:queryBall.objects){
          KNNList knnlist2 = aknnJoinResult.get(temp.getID());
          if(knnlist2 != null){
            globalKnnDist = Math.max(knnlist2.getKNNdist(), globalKnnDist);
          }else{
            globalKnnDist = Double.MAX_VALUE;
            break;
          }
        }
      }
      
    }
    return aknnJoinResult;
  }

  
  double mindist(DoubleVector dv, SpatialEntry se){
    double mindist = 0;
    for (int i = 0; i < dv.getDimensionality(); i++) {
      double q = dv.getValue(i + 1);
      double lb = se.getMin(i + 1);
      double ub = se.getMax(i + 1);
      if (q < lb) {
        mindist += (lb - q) * (lb - q);
      } else if (q > ub) {
        mindist += (q - ub) * (q - ub);
      }
    }
    return Math.sqrt(mindist);
  }
}
