package experimentalcode.tobias.TrigonometricPruning;

import java.util.List;

import de.lmu.ifi.dbs.elki.distance.distancefunction.EuclideanDistanceFunction;
import experimentalcode.tobias.TrigonometricPruning.structures.DataBall;
import experimentalcode.tobias._index.sstree.DoubleVector;
import experimentalcode.tobias._index.sstree.NodeEntry;


public class PruningHeuristic {
  
  public static int distanceCalculations = 0;
  
  static EuclideanDistanceFunction edf = EuclideanDistanceFunction.STATIC;
  
  public static boolean canBePrunedCos(DoubleVector dv, DataBall queryBall, double mindist, List<PruningCandidate<DoubleVector>> pruningCandidates, int k, double kMaxPCDist){
    
    double dvToCenter = mindist + queryBall.radius;
    int pruneCount = 0;
//    only check if we have k pruningCandidates
    if(pruningCandidates.size()>=k){
      if(dvToCenter+queryBall.radius<kMaxPCDist){
        ;
      }else if(dvToCenter-queryBall.radius > kMaxPCDist){
        pruneCount = k;
      }else{
        for(int i = 0; i< pruningCandidates.size();i++){
          PruningCandidate<DoubleVector> pc = pruningCandidates.get(i);
          double pruningCandidateDist = pc.centerDist;
          double betweenDist = edf.distance(dv, pc.pruningCandidate).doubleValue();
          distanceCalculations+=1;
          
          
          double distToLine = -(pruningCandidateDist*pruningCandidateDist-dvToCenter*dvToCenter)/(2*betweenDist);
          if(distToLine < queryBall.radius)
            continue;
          
          pruneCount++;
          if(pruneCount >= k)break;
        }
      }
    }
    
    
    if(pruneCount<k){
      //pruningCandidates.add(new PruningCandidate<DoubleVector>(dvToCenter, dv));
      return false;
    }
  return true;
}
public static boolean canBePrunedCos(NodeEntry page, DataBall queryBall, double mindist, List<PruningCandidate<DoubleVector>> pruningCandidates, int k, double kMaxPCDist){
  
  double centerToCenter = mindist + page.radius + queryBall.radius;
    
  int pruneCount = 0;
  if(pruningCandidates.size()>=k){
    if(centerToCenter-page.radius<kMaxPCDist){
      ;
    }else{
      for(int i = 0; i < pruningCandidates.size();i++){
        PruningCandidate<DoubleVector> pc = pruningCandidates.get(i);
        double betweenDist = edf.distance(page.centroid, pc.pruningCandidate).doubleValue();
        distanceCalculations++;
        double candidateDist = centerToCenter;
        //a^2 = b^2+c^2-2bc*cos(alpha)
        double delta = Math.acos((betweenDist*betweenDist-candidateDist*candidateDist-pc.centerDist*pc.centerDist)/(-2*candidateDist*pc.centerDist));
        double temp = -(candidateDist*Math.sin(delta)/(pc.centerDist-candidateDist*Math.cos(delta)));
        double alpha = Math.atan(temp);
        
        double d1 = Math.sqrt(pc.centerDist*pc.centerDist+queryBall.radius*queryBall.radius-2*pc.centerDist*queryBall.radius*Math.cos(alpha));
        double d2 = Math.sqrt(candidateDist*candidateDist+queryBall.radius*queryBall.radius-2*candidateDist*queryBall.radius*Math.cos(alpha-delta))-page.radius;
        if(d2 < d1){
          continue;
        }
        
        alpha = alpha+Math.PI;
        d1 = Math.sqrt(pc.centerDist*pc.centerDist+queryBall.radius*queryBall.radius-2*pc.centerDist*queryBall.radius*Math.cos(alpha));
        d2 = Math.sqrt(candidateDist*candidateDist+queryBall.radius*queryBall.radius-2*candidateDist*queryBall.radius*Math.cos(alpha-delta))-page.radius;
        if(d2 < d1)continue;
  
        pruneCount++;
        if(pruneCount >= k){
          break;
        }
      }
    }
  }
  if(pruneCount<k){
    
    return false;
  }
  return true;
  
}
  public static class PruningCandidate<T> implements Comparable<PruningCandidate<T>>{
    public double centerDist;
    public T pruningCandidate;
    public PruningCandidate(double centerDist, T pruningCandidate){
      this.centerDist = centerDist;
      this.pruningCandidate = pruningCandidate;
    }
    @Override
    public int compareTo(PruningCandidate<T> o) {
     if(centerDist < o.centerDist)return -1;
     else if(o.centerDist < centerDist)return 1;
     else return 0;
    }
  }
}
