package experimentalcode.tobias.TrigonometricPruning;

import java.util.ArrayList;
import java.util.List;
import java.util.PriorityQueue;
import java.util.TreeMap;

import de.lmu.ifi.dbs.elki.database.ids.DBID;
import de.lmu.ifi.dbs.elki.distance.distancefunction.EuclideanDistanceFunction;
import experimentalcode.tobias.TrigonometricPruning.PruningHeuristic.PruningCandidate;
import experimentalcode.tobias.TrigonometricPruning.structures.DataBall;
import experimentalcode.tobias._index.helpstructures.KNNList;
import experimentalcode.tobias._index.sstree.DoubleVector;
import experimentalcode.tobias._index.sstree.NodeEntry;
import experimentalcode.tobias._index.sstree.SSDataNode;
import experimentalcode.tobias._index.sstree.SSDirectoryNode;
import experimentalcode.tobias._index.sstree.SSNode;
import experimentalcode.tobias._index.sstree.SSTree;
import experimentalcode.tobias.rknnjoin.structures.ObjectScorePair;
import experimentalcode.tobias.rknnjoin.structures.buffer.Buffer2;


//Join according to zhang, mamoulis, papadias, tao paper on an SSTree
public class SSBNN extends AkNNJoin{
  
  public static final int BNN = 0, MAX = 1, TRIG = 2, TRIGBNN = 3;
  public static final String[] type = new String[]{"BNN", "MAX", "TP", "TPBNN"}; 
  int pruning;
  int epsilon=2;
  
  SSTree<DoubleVector> ssTree;

  Buffer2<SSNode> lruBuffer;
  EuclideanDistanceFunction edf;
  public long time;
  
  public SSBNN(SSTree<DoubleVector> ssTree, int bufferSize){
    this.ssTree = ssTree;
    edf = EuclideanDistanceFunction.STATIC;
    lruBuffer = new Buffer2<SSNode>(bufferSize);
  }

  @Override
  public TreeMap<DBID,KNNList<DoubleVector>> join(DataBall queryBall, int k){
    TreeMap<DBID,KNNList<DoubleVector>> aknnJoinResult = new TreeMap<DBID,KNNList<DoubleVector>>();
    
    KNNList<PruningCandidate<DoubleVector>> pruningCandidates = new KNNList<PruningCandidate<DoubleVector>>(epsilon*k);
    boolean pcChanged = false;
    KNNList maxDist = new KNNList(k);
    
    //Organize pages by mindist
    PriorityQueue<ObjectScorePair<NodeEntry>> apl = new PriorityQueue<ObjectScorePair<NodeEntry>>();
    for(NodeEntry ne:((SSDirectoryNode<DoubleVector>)ssTree.root).getNodeEntrys()){
      double dist = edf.distance(queryBall.center, ne.centroid).doubleValue()-ne.radius;
      distanceCalculations++;
      apl.add(new ObjectScorePair(ne, dist));
    }
    double globalKnnDist = Double.MAX_VALUE;
    while(!apl.isEmpty()){
      double minDist = apl.peek().score;
      if(globalKnnDist < minDist)break;
      
      NodeEntry nodeEntry = apl.poll().object;
      //MaxDistPruning
      if(pruning >= MAX && minDist> maxDist.getKNNdist())continue;
      if((pruning >= TRIG)){
        if(PruningHeuristic.canBePrunedCos(nodeEntry, queryBall,minDist, pruningCandidates.getKNNs(), k, maxDist.getKNNdist())){
          continue;
        }
      }
      
      SSNode ssnode = nodeEntry.ssnode;
      
      logicalPageAccesses++;
      if(lruBuffer.add(ssnode))
        physicalPageAccesses++;
      
      
      
      if(ssnode instanceof SSDirectoryNode){
        for(NodeEntry ne:((SSDirectoryNode<DoubleVector>)ssnode).getNodeEntrys()){
          double dist = edf.distance(queryBall.center, ne.centroid).doubleValue()-ne.radius-queryBall.radius;
          distanceCalculations++;
          apl.add(new ObjectScorePair(ne, dist));
        }
      }else{
      //Remove all query vectors which can not be updated
        List<DoubleVector> updateQuery = new ArrayList<DoubleVector>();
        for(DoubleVector dv:queryBall.objects){
          KNNList knnlist = aknnJoinResult.get(dv.getID());
          if(knnlist != null){
            distanceCalculations++;
            if(knnlist.getKNNdist() < edf.distance(dv, ssnode.accomdatingEntry.centroid).doubleValue()-ssnode.accomdatingEntry.radius)
                continue; 
          }
          updateQuery.add(dv);
        }
        //Remove all data vectors which can not contribute new knns
        List<DoubleVector> kNNcandidates = new ArrayList<DoubleVector>();
        for(DoubleVector dvs:((SSDataNode<DoubleVector>)ssnode).getData()){
          distanceCalculations++;
          double dist = edf.distance(dvs, queryBall.center).doubleValue();
          
          
          if((pruning == BNN || pruning == TRIGBNN) && dist-queryBall.radius>globalKnnDist){
            ;
          }/*else if(pruning >= TRIG && PruningHeuristic.canBePrunedCos(dvs, queryBall,dist - queryBall.radius, pruningCandidates, k, maxDist.getKNNdist())){
            ;
          }*/else{
            kNNcandidates.add(dvs);
            pruningCandidates.add(new PruningCandidate<DoubleVector>(dist, dvs), dist);
          }
          maxDist.add(dvs, dist+queryBall.radius);
        }
        
        
        for(DoubleVector dv:updateQuery){
          for(DoubleVector dvs:kNNcandidates){
            addResult(dv, dvs, k, aknnJoinResult);
            
          }
        }
        
        if(pruning == BNN || pruning == TRIGBNN){
        //update global KNNDist
          globalKnnDist = 0.0;
          for(DoubleVector temp:queryBall.objects){
            KNNList knnlist2 = aknnJoinResult.get(temp.getID());
            if(knnlist2 != null){
              globalKnnDist = Math.max(knnlist2.getKNNdist(), globalKnnDist);
            }else{
              globalKnnDist = Double.MAX_VALUE;
              break;
            }
          }
        }
      /*  long t = System.nanoTime();
        if(pruning >= TRIG && pcChanged){
          //update pruningCandidates
          Collections.sort(pruningCandidates);
          int cutoff = Integer.MAX_VALUE;
          for(int i =0 ; i< pruningCandidates.size();i++){
            if(pruningCandidates.get(i).centerDist + queryBall.radius > maxDist.getKNNdist()* 2*queryBall.radius){
              cutoff = i;break;
            }
          }
          if(cutoff>epsilon*k)
            cutoff = epsilon*k;
          if(cutoff < pruningCandidates.size())
            pruningCandidates = pruningCandidates.subList(0, cutoff+1);
        }
        time += System.nanoTime()-t;*/
      }
      
    }
    distanceCalculations += PruningHeuristic.distanceCalculations;
    PruningHeuristic.distanceCalculations =0;
    //System.out.println(pruningCandidates.size());
    return aknnJoinResult;
  }
  public void setPruning(int p){
    pruning = p;
  }
  
  public void reset(){
    distanceCalculations = 0;
    logicalPageAccesses = 0;
    physicalPageAccesses = 0;
    lruBuffer.clear();
    
  }
}
