package experimentalcode.tobias.TrigonometricPruning;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.List;
import java.util.TreeMap;

import de.lmu.ifi.dbs.elki.database.ids.DBID;
import de.lmu.ifi.dbs.elki.index.tree.spatial.rstarvariants.rstar.RStarTree;
import de.lmu.ifi.dbs.elki.utilities.optionhandling.ParameterException;
import experimentalcode.tobias.TrigonometricPruning.structures.DataBall;
import experimentalcode.tobias.TrigonometricPruning.utils.Converter;
import experimentalcode.tobias._index.helpstructures.KNNList;
import experimentalcode.tobias._index.sstree.DoubleVector;
import experimentalcode.tobias._index.sstree.SSTree;
import experimentalcode.tobias.rknnjoin.utils.CSVutils;

public class TrigPruningTest {

  /**
   * @param args
   * @throws IOException 
   * @throws NumberFormatException 
   * @throws ParameterException 
   */
  public static void main(String[] args) throws NumberFormatException, IOException, ParameterException {
    
    String dataset = "../data/csv/forest10.csv";
    String resultFile = "../data/csv/results.txt";
    int pageSize = 4096; //Byte
    int innerSet = 387342;
    int outerSet = 193670;
    int bufferSize = 10; //pages
    int k = 10;
    
    int epsilon = 2;
    
    BufferedWriter out = new BufferedWriter(new FileWriter(resultFile, true));
    String settings = "dataset = "+dataset+"\n";
    settings += "pageSie = "+pageSize+"\n";
    settings += "innerSet = "+innerSet+"\n";
    settings += "outerSet = "+outerSet+"\n";
    settings += "bufferSize = "+bufferSize+"\n";
    settings += "k = "+k+"\n";
    out.write(settings);

    
    System.out.println("Reading Data...");
    List<DoubleVector> data = CSVutils.readCSV(dataset, -1, innerSet+outerSet);
    List<DoubleVector> innerData = data.subList(0, innerSet);
    List<DoubleVector> outerData = data.subList(innerSet, data.size());
    int dimensionality = data.get(0).getDimensionality();
    int M = pageSize / (12+dimensionality*8);
    int m = (int)(0.4 * M);
    
    System.out.println("BuildingIndexes");
    //RTree
    RStarTree<DoubleVector> rTree = new RStarTree(null, null, pageSize, Integer.MAX_VALUE, false, null, 0);
    rTree.insert(innerData);
  //SSTree
    SSTree ssTree = new SSTree(m, M,0);
    ssTree.insert(innerData);
    
    double dataBallRadius = 1.2*ssTree.avgDNRad();
    
    AkNNJoin bnn = new BNN(rTree, bufferSize);
    AkNNJoin ssbnn = new SSBNN(ssTree, bufferSize);
    ((SSBNN)ssbnn).epsilon = epsilon;
    
    System.out.println("Making QueryBalls");
    List<DataBall> queryBalls = Converter.makeDataBalls(outerData, dataBallRadius);
    
    TreeMap<DBID, KNNList<DoubleVector>> ssResult = new TreeMap<DBID, KNNList<DoubleVector>>();
    TreeMap<DBID, KNNList<DoubleVector>> rResult = new TreeMap<DBID, KNNList<DoubleVector>>();
    
    //RTree BNN
    for(int i = 0; i< queryBalls.size();i++){
      //System.out.println("Joining Ball "+i+"/"+queryBalls.size());
      rResult.putAll(bnn.join(queryBalls.get(i), k));
    }
    System.out.println("BNN:\n "+bnn);
    out.write("BNN:\n "+bnn);

    for(int i = 0; i< 4; i++){
      ((SSBNN)ssbnn).setPruning(i);
      long t = System.currentTimeMillis();
      for(int j = 0; j< queryBalls.size();j++){
       // System.out.println("Joining Ball "+j+"/"+queryBalls.size());
        ssResult.putAll(ssbnn.join(queryBalls.get(j), k));
      }
      System.out.println("SSBNN("+SSBNN.type[i]+"):\n "+ssbnn);
      System.out.println("Took "+(System.currentTimeMillis()-t)+" ms");
      out.write("SSBNN("+SSBNN.type[i]+"):\n "+ssbnn);
      out.write("\n Took "+(System.currentTimeMillis()-t)+" ms");
      ((SSBNN)ssbnn).reset();
      ssResult.clear();
    }
    //Compare.compareResults(Compare.bruteForceKnn(outerData, innerData, k),ssResult);
    /*System.out.println(rTree);
    System.out.println(ssTree);
    System.out.println("BNN:\n "+bnn);
    System.out.println("SSBNN:\n "+ssbnn);*/
    out.close();


  }

}
