package experimentalcode.tobias.TrigonometricPruning.utils;

import java.util.ArrayList;
import java.util.List;

import de.lmu.ifi.dbs.elki.distance.distancefunction.EuclideanDistanceFunction;

import experimentalcode.tobias.TrigonometricPruning.structures.DataBall;
import experimentalcode.tobias._index.sstree.DoubleVector;
import experimentalcode.tobias._index.sstree.NodeEntry;
import experimentalcode.tobias._index.sstree.SSDataNode;
import experimentalcode.tobias._index.sstree.SSDirectoryNode;
import experimentalcode.tobias._index.sstree.SSNode;
import experimentalcode.tobias._index.sstree.SSTree;



public class Converter {
  static EuclideanDistanceFunction edf = EuclideanDistanceFunction.STATIC;
  
	/**
	 * Takes an sstree and return only the leaf nodes as a list
	 * 
	 * @param sstree
	 * @return
	 */
	public static List<DataBall> makeDataBalls(SSTree sstree) {
		List<DataBall> dataBalls = new ArrayList<DataBall>();
		List<SSNode> nodes = new ArrayList<SSNode>();
		nodes.add(sstree.root);
		while (!nodes.isEmpty()) {
			SSNode node = nodes.remove(0);
			if (node instanceof SSDirectoryNode) {
				for (NodeEntry ne : ((SSDirectoryNode<DoubleVector>) node).getNodeEntrys()) {
					nodes.add(ne.ssnode);
				}
			} else {
				SSDataNode da = (SSDataNode) node;
				dataBalls.add(new DataBall(da.getData(),
						da.accomdatingEntry.centroid,
						da.accomdatingEntry.radius));
			}
		}
		return dataBalls;
	}

	/**
	 * Clusters datapoints in a birch-like approach
	 */
	public static List<DataBall> makeDataBalls(List<DoubleVector> db,
			double maxRadius) {
		List<DataBall> dataBalls = new ArrayList<DataBall>();
		DoubleVector dv = db.get(0);
		List<DoubleVector> dataBallList = new ArrayList<DoubleVector>();
		dataBallList.add(dv);
		dataBalls.add(new DataBall(dataBallList, dv));

		for (int j = 1; j<db.size(); j++) {
			dv = db.get(j);
			// look for the best ball
			int bestBall = 0;
			double minDist = edf.distance(dataBalls.get(0).center, dv).getValue();
			for (int i = 1; i < dataBalls.size(); i++) {
				if (edf.distance(dataBalls.get(i).center, dv).getValue() < minDist) {
					bestBall = i;
					minDist = edf.distance(dataBalls.get(i).center, dv).getValue();
				}
			}
			// put point in the best ball
			dataBalls.get(bestBall).objects.add(dv);
			DoubleVector newCentroid = calcCentroid(dataBalls.get(bestBall).objects);
			double newRadius = calcMaxDist(dataBalls.get(bestBall).objects,
					newCentroid);
			if (newRadius < maxRadius) {
				dataBalls.get(bestBall).center = newCentroid;
				dataBalls.get(bestBall).radius = newRadius;
			} else {// we have to split the balls
				List<DoubleVector> fvs = dataBalls.get(bestBall).objects;
				// look for the two points with the biggest distance
				DoubleVector maxa = null, maxb = null;
				double maxDist = 0;
				for (int a = 0; a < fvs.size(); a++) {
					for (int b = 0; b < fvs.size(); b++) {
						double dist = edf.distance(fvs.get(a), fvs.get(b)).getValue();
						if (dist > maxDist) {
							maxDist = dist;
							maxa = fvs.get(a);
							maxb = fvs.get(b);
						}
					}
				}
				DataBall oldBall = dataBalls.remove(bestBall);
				DataBall newBall1 = new DataBall(
						new ArrayList<DoubleVector>(), maxa);
				DataBall newBall2 = new DataBall(
						new ArrayList<DoubleVector>(), maxb);
				for (DoubleVector f : oldBall.objects) {
					double dist1 = edf.distance(f, newBall1.center).getValue();
					double dist2 = edf.distance(f, newBall2.center).getValue();
					if (dist1 < dist2)
						newBall1.objects.add(f);
					else
						newBall2.objects.add(f);
				}
				newBall1.center = calcCentroid(newBall1.objects);
				newBall2.center = calcCentroid(newBall2.objects);
				newBall1.radius = calcMaxDist(newBall1.objects, newBall1.center);
				newBall2.radius = calcMaxDist(newBall2.objects, newBall2.center);
				dataBalls.add(newBall1);
				dataBalls.add(newBall2);
			}
		}

		return dataBalls;
	}
	/**
	 * Clusters datapoints with a hilbert curve
	 */
/*	public static List<DataBall> makeDataBalls(Database<DoubleVector> db, double maxArea, boolean hilbert){
		DistanceMeasure dm = new EuclidianDistance();
		List<DataBall> dataBalls = new ArrayList<DataBall>();
		System.out.println("Clustering: "+maxArea);
		List<MBRList<DoubleVector>> cl1 = (new HilbertClusterer<DoubleVector>(db, maxArea, 2000, 3)).getClusters(); 
		System.out.println("Converting");
		for(MBRList mbrlist: cl1){
			ArrayList<DoubleVector> dataList = new ArrayList<DoubleVector>();
			for(DoubleVector fv: (MBRList<DoubleVector>) mbrlist){
				dataList.add(fv);
			}
			dataBalls.add(new DataBall(dataList));
		}
		System.out.println("Shrinking");
		shrinkDataBalls(dataBalls);
		return dataBalls;
	}*/
	/**
	 * Clusters datapoints in a birch-like approach and returns the result as a
	 * list of MBRClusters.
	 */
	/*public static List<MBRList<DoubleVector>> makeDataMBRs(Database<DoubleVector> db,
			double maxRadius) {
		List<DataBall> dbs = makeDataBalls(db, maxRadius);
		List<MBRList<DoubleVector>> ret = new ArrayList<MBRList<DoubleVector>>();
		for (Iterator<DataBall> iterator = dbs.iterator(); iterator.hasNext();) {
			DataBall list = iterator.next();
			ret.add(new MBRList<DoubleVector>(list.objects));
		}
		return ret;
	}*/

	private static DoubleVector calcCentroid(List<DoubleVector> fvs) {
		double[] centroid = new double[fvs.get(0).getDimensionality()];
		for (DoubleVector fv : fvs) {
			for (int d = 0; d < centroid.length; d++) {
				centroid[d] += fv.getValue(d+1);
			}
		}
		for (int d = 0; d < centroid.length; d++)
			centroid[d] /= fvs.size();
		return new DoubleVector(centroid);
	}

	protected static double calcMaxDist(List<DoubleVector> fvs,
			DoubleVector centroid) {
		double maxRad = 0.0;

		for (DoubleVector fv : fvs)
			maxRad = Math.max(maxRad, edf.distance(fv,
					centroid).getValue());
		return maxRad;
	}

	/**
	 * computes the approximate minimum bounding sphere for each data point
	 */
	public static void shrinkDataBalls(List<DataBall> dataBalls) {
		double ratio = 0.0;
		for (DataBall dataBall : dataBalls) {
			SEB seb = SEBcomputation.getApproximateSEB(dataBall.objects, 0.01);
			double oldRadius = dataBall.radius;
			dataBall.radius = seb.radius;
			dataBall.center = seb.center;
			ratio += seb.radius / oldRadius;
		}
		// System.out.println("Shrank balls about "+(1.0-ratio/dataBalls.size())*100.0+"%");
	}

}
