package experimentalcode.tobias.TrigonometricPruning.utils;


import java.util.ArrayList;
import java.util.List;

import de.lmu.ifi.dbs.elki.distance.distancefunction.EuclideanDistanceFunction;
import experimentalcode.tobias._index.sstree.DoubleVector;

public class SEBcomputation {
  static EuclideanDistanceFunction edf = EuclideanDistanceFunction.STATIC;

	/**
	 * returns the approximate center and radius of the samallest enclosing ball. Radius of the approximate SEB is
	 * (1+epsilon)*Radius of the real SEB. Runtime of this Algorithm is O(d*n/e^2)
	 * @param points the n points
	 * @param epsilon approximation variable
	 * @return
	 */
	public static SEB getApproximateSEB(List<DoubleVector> points, double epsilon){
	  
		ArrayList<DoubleVector> centerSequence = new ArrayList<DoubleVector>();
		centerSequence.add(points.get(0));
		for(int i = 1; i<= (1.0/(epsilon*epsilon));i++){
		  DoubleVector q = null;
			double maxDist = 0.0;
			for(DoubleVector fv:points){
				double dist = edf.distance(fv, centerSequence.get(i-1)).getValue();
				if(dist >= maxDist){
					maxDist = dist;
					q = fv;
				}
			}
			centerSequence.add(Miscellaneous.plus(centerSequence.get(i-1), Miscellaneous.scalar(Miscellaneous.minus(q,centerSequence.get(i-1)), 1.0/i)));
		}
		
		DoubleVector center = centerSequence.get(centerSequence.size()-1);
		return new SEB(center, getRadius(points, center));
	}
	public static double getRadius(List<DoubleVector> points, DoubleVector center){
		double maxDist = 0.0;
		for(DoubleVector fv:points){
			maxDist = Math.max(maxDist, edf.distance(fv, center).getValue());
		}
		return maxDist;
	}
	
}
