package experimentalcode.tobias._index.sstree;


import java.util.ArrayList;
import java.util.List;
import java.util.PriorityQueue;



public class SSDirectoryNode<T extends DoubleVector> extends SSNode<T>{
	
	private List<NodeEntry> children;
	
	public SSDirectoryNode(SSTree sstree, boolean isRoot, SSDirectoryNode parent, int m, int M, NodeEntry accomdatingEntry){
		super(sstree, isRoot, parent, m, M, accomdatingEntry);
		children = new ArrayList<NodeEntry>();
	}

	@Override
	public void insert(T fv) {
		double minDistToCentroid = Double.MAX_VALUE;
		int indexOfChild = -1;
		//Find the clothest centroid among the childpages
		for(int c = 0; c<children.size();c++){
			double dist = df.distance(children.get(c).centroid, fv).getValue();
			if(dist<minDistToCentroid){
				minDistToCentroid = dist;
				indexOfChild = c;
			}
		}
		children.get(indexOfChild).ssnode.insert(fv);
	}

	@Override
	public void update(boolean recursive) {
		if(children.size() > M){
			split();
		}else{
			int tData = 0;
			ArrayList<T> childCentroids = new ArrayList<T>();
			for(NodeEntry ne:children){
				childCentroids.add((T)ne.centroid);
				tData+=ne.totalDataPoints;
			}
			DoubleVector centroid = calcCentroid(childCentroids);
			
			double radius = 0.0;
			for(NodeEntry ne:children){
				radius = Math.max(radius, df.distance(centroid, ne.centroid).getValue()+ne.radius);
			}
			
			accomdatingEntry.totalDataPoints = tData;
			accomdatingEntry.updateCR(centroid, radius);
			if(!isRoot()&&recursive)parent.update(true);
			
			
		}
	}
	/**
	 * finds the dimension with the highest variance
	 *
	 */
	private void split(){
		//1. Find the splitting dimension
		ArrayList<T> toSplit = new ArrayList<T>();
		for(NodeEntry childNodeCentroid:children)
			toSplit.add((T)childNodeCentroid.centroid);
		int maxVarDim = calcSplitAxis(toSplit);
		//2. Sort centroids according to split-dimension
		PriorityQueue<DistanceObjectTuple> nodeEntryQueue = new PriorityQueue<DistanceObjectTuple>();
		for(NodeEntry child:children)
			nodeEntryQueue.add(new DistanceObjectTuple(child.centroid.getValue(maxVarDim+1), child));
		ArrayList<DoubleVector> sortedCentroids = new ArrayList<DoubleVector>();
		ArrayList<NodeEntry> sortedEntryList = new ArrayList<NodeEntry>();
		while(!nodeEntryQueue.isEmpty()){
			NodeEntry ne = (NodeEntry)(nodeEntryQueue.poll().object);
			sortedCentroids.add(ne.centroid);
			sortedEntryList.add(ne);
		}
		//3. Find the best splitting Index
		int bestSplitIndex = splitIndex(maxVarDim, (List<T>)sortedCentroids);
		//4. Then split according to the best splitting found
		List<NodeEntry> node1 = getSubList(0, bestSplitIndex, sortedEntryList);
		List<NodeEntry> node2 = getSubList(bestSplitIndex, sortedEntryList.size(), sortedEntryList);
		
		if(isRoot()){
			//this node stays root
			this.children.clear();
			NodeEntry child1 = addNodeEntry();
			NodeEntry child2 = addNodeEntry();
			SSDirectoryNode ssdnode1 = new SSDirectoryNode(sstree, false, this, m, M, child1);
			SSDirectoryNode ssdnode2 = new SSDirectoryNode(sstree, false, this, m, M, child2);
			ssdnode1.children = node1;
			ssdnode2.children = node2;
			
			for(NodeEntry ne: node1){
				ne.ssnode.parent = ssdnode1;
			}
			for(NodeEntry ne: node2){
				ne.ssnode.parent = ssdnode2;

			}

			child1.setSSNode(ssdnode1);
			child2.setSSNode(ssdnode2);
			
			ssdnode1.update(false);
			ssdnode2.update(false);
			this.update(true);
			
		}else{
			//remove data from old Node
			children = node1;
			//put the removed data in a new node
			NodeEntry newEntry = parent.addNodeEntry();
			SSDirectoryNode ssdnode = new SSDirectoryNode(sstree, false, parent, m, M, newEntry);
			ssdnode.children = node2;
			for(NodeEntry ne: node2)
				ne.ssnode.parent = ssdnode;
			newEntry.setSSNode(ssdnode);
			
			//recalculate the approximations
			this.update(false);
			ssdnode.update(false);
			parent.update(true);
		}
	}
	public List<NodeEntry> getNodeEntrys(){
		return children;
	}
	@Override
  public int size(){
		return children.size();
	}
	public NodeEntry addNodeEntry(){
		NodeEntry newNodeEntry = new NodeEntry();
		children.add(newNodeEntry);
		return newNodeEntry;
	}
}
