package experimentalcode.tobias._index.sstree;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.PriorityQueue;
import java.util.TreeMap;

import de.lmu.ifi.dbs.elki.database.ids.DBID;
import de.lmu.ifi.dbs.elki.distance.distancefunction.EuclideanDistanceFunction;
import experimentalcode.tobias.rknnjoin.structures.buffer.Buffer2;

public class SSTree <T extends DoubleVector>{

	int m,M;
	public SSNode<T> root;
	public int distanceCalculations = 0, physicalPageAccesses = 0, logicalPageAccesses = 0;
	private TreeMap<DBID, T> keyMap = new TreeMap<DBID, T>();
	EuclideanDistanceFunction df;
	Buffer2 lruBuffer;
	
	
    PriorityQueue<DistanceObjectTuple> incrementalApl;
    
	public SSTree(int m, int M, int bufferSize) {
		df = EuclideanDistanceFunction.STATIC;
		this.m = m;
		this.M=M;
		root = new SSDataNode(this, true,null, m, M, new NodeEntry());
		root.accomdatingEntry.ssnode = root;
		lruBuffer = lruBuffer = new Buffer2(bufferSize);
	}
/**
 * returns an unordered List of double vectors which lie within the epsilon range of query
 * @param query
 * @param range
 * @return
 */
	public List<T> epsRange(T query, double range) {
		PriorityQueue<DistanceObjectTuple> apl = new PriorityQueue<DistanceObjectTuple>();
		ArrayList result = new ArrayList();
		apl.add(new DistanceObjectTuple(0.0, root));
		while(!apl.isEmpty() && apl.peek().distance < range){
			SSNode currentNode = (SSNode)apl.poll().object;
			
			logicalPageAccesses++;
			if(lruBuffer.add(currentNode))
			  physicalPageAccesses++;
			  
			if(currentNode instanceof SSDataNode){
				for(DoubleVector dv:((SSDataNode<T>)currentNode).getData()){
					double dist = df.distance(dv, query).getValue();
					distanceCalculations++;
					if(dist < range)
						result.add(dv);
				}
					
			}else{
				for(NodeEntry ne:((SSDirectoryNode<T>)currentNode).getNodeEntrys()){
					double dist = df.distance(query, ne.centroid).getValue()-ne.radius;
					distanceCalculations++;
					if(dist < range){
						apl.add(new DistanceObjectTuple(dist, ne.ssnode));
					}
				}
			}
		}
		return result;
	}

	public int getCount() {
		return totalData(this.root);
	}

	public T getInstance(String key) {
		return keyMap.get(key);
	}

	public void getInstanceSample(T[] sample) {
		throw new UnsupportedOperationException();
	}

	public T getNext(T query, double maxDist, double[] distance) {
		T result = getNext(query);
		distance[0] = df.distance(query, result).getValue();
		return result;
	}

	public List<T> getStratifiedSample(int size) {
		throw new UnsupportedOperationException();
	}

	public DBID insert(T newInstance) {
		keyMap.put(newInstance.getID(), newInstance);
		DBID pK = newInstance.getID();
		root.insert(newInstance);
		return pK;
	}
	public void insert(List<T> data) {
	  for(T t:data)
	    insert(t);
  }
	public List<T> kNNQuery(T query, int k) {
		PriorityQueue<DistanceObjectTuple> apl = new PriorityQueue<DistanceObjectTuple>();
		PriorityQueue<DistanceObjectTuple> knn = new PriorityQueue<DistanceObjectTuple>(100,new DistanceObjectTuple.DesendingComparator());
		apl.add(new DistanceObjectTuple(0, root));
		while(!apl.isEmpty() && (knn.size()<k || apl.peek().distance < knn.peek().distance)){
			SSNode currentNode = (SSNode)apl.poll().object;
			
			logicalPageAccesses++;
      if(lruBuffer.add(currentNode))
        physicalPageAccesses++;
      
			if(currentNode instanceof SSDataNode){
				for(T dv:((SSDataNode<T>)currentNode).getData()){
					knn.add(new DistanceObjectTuple(df.distance(dv, query).getValue(), dv));
					distanceCalculations++;
					if(knn.size()>k)knn.poll();
				}
					
			}else{
				for(NodeEntry ne:((SSDirectoryNode<T>)currentNode).getNodeEntrys()){
					double dist = df.distance(query, ne.centroid).getValue()-ne.radius;
					distanceCalculations++;
					if(knn.size()<k || dist < knn.peek().distance){
						apl.add(new DistanceObjectTuple(dist, ne.ssnode));
					}
				}
			}
		}
		List<T> result = new ArrayList<T>();
		while(!knn.isEmpty()){
			result.add(0,(T)knn.poll().object);
		}
		//System.out.println("Distance-Calculations: "+distanceCalcs + " Page-Accesses: "+pageAccesses);
		return result;
	}


	public Iterator<T> objectIterator() {
		return keyMap.values().iterator();
	}


	public void reset(){
		incrementalApl = new PriorityQueue();
		incrementalApl.add(new DistanceObjectTuple(0, root));
		distanceCalculations = 0;
		logicalPageAccesses=0;
		physicalPageAccesses=0;
		lruBuffer.clear();
	}
	public T getNext(T query){
		if(incrementalApl == null)reset();
		while(!incrementalApl.isEmpty()){
			Object o = incrementalApl.poll().object;
			if(o instanceof DoubleVector)
				return (T)o;
			else{
				SSNode currentNode = (SSNode)o;
				
				logicalPageAccesses++;
	      if(lruBuffer.add(currentNode))
	        physicalPageAccesses++;
	      
				if(currentNode instanceof SSDataNode){
					for(T dv:((SSDataNode<T>)currentNode).getData()){
						incrementalApl.add(new DistanceObjectTuple(df.distance(dv, query).getValue(), dv));
						distanceCalculations++;
					}
						
				}else{
					for(NodeEntry ne:((SSDirectoryNode<T>)currentNode).getNodeEntrys()){
						double dist = df.distance(query, ne.centroid).getValue()-ne.radius;
						distanceCalculations++;
						incrementalApl.add(new DistanceObjectTuple(dist, ne.ssnode));
					}
				}
			}
		}
		return null;
	}
	public double getNextMinDist(){
		return incrementalApl.peek().distance;
	}
	public String toString(SSNode node, int space){
		String out = "";
		String freeSpace = "";
		for(int i = 0; i< space;i++)freeSpace+="-";
		out += freeSpace + node.toString();
		if(node instanceof SSDirectoryNode){
			List<NodeEntry> l = ((SSDirectoryNode)node).getNodeEntrys();
			for(NodeEntry ne:l)
				out += "\n"+toString(ne.ssnode, space+1);
		}else{
			List<T> l = ((SSDataNode)node).getData();
			for(T fv:l)
				out += "\n-"+freeSpace+fv.getID();
		}
		return out;
	}
	public int totalNodes(){
		return totalNodes(this.root);
	}
	private int totalNodes(SSNode node){
		int count = 1;
		if(node instanceof SSDirectoryNode){
			for(NodeEntry ne:((SSDirectoryNode<T>)node).getNodeEntrys()){
				count+=totalNodes(ne.ssnode);
			}
		}
		return count;
	}
	private int totalData(SSNode node){
		int count = 0;
		if(node instanceof SSDirectoryNode){
			for(NodeEntry ne:((SSDirectoryNode<T>)node).getNodeEntrys()){
				count+=totalData(ne.ssnode);
			}
		}else{
			count = ((SSDataNode)node).size();
		}
		return count;
	}
	@Override
  public String toString(){
		int directoryNodeCount = 0;
		int dataNodeCount = 0;
		double avgRad = 0.0;
		ArrayList<SSNode> toView = new ArrayList<SSNode>();
		toView.add(root);
		while(!toView.isEmpty()){
			SSNode node = toView.remove(0);
			if(node instanceof SSDataNode){
				dataNodeCount++;
				avgRad+=node.accomdatingEntry.radius;
			}else{
				directoryNodeCount++;
				for(NodeEntry ne:((SSDirectoryNode<T>)node).getNodeEntrys()){
					toView.add(ne.ssnode);
				}
			}
		}
		return "Directory-Nodes: "+directoryNodeCount+ " Data-Nodes: "+dataNodeCount+" AVG-DN-Rad: "+(avgRad/dataNodeCount);
	}
	public double avgDNRad(){
		int dataNodeCount = 0;
		double avgRad = 0.0;
		ArrayList<SSNode> toView = new ArrayList<SSNode>();
		toView.add(root);
		while(!toView.isEmpty()){
			SSNode node = toView.remove(0);
			if(node instanceof SSDataNode){
				dataNodeCount++;
				avgRad+=node.accomdatingEntry.radius;
			}else{
				for(NodeEntry ne:((SSDirectoryNode<T>)node).getNodeEntrys()){
					toView.add(ne.ssnode);
				}
			}
		}
		return (avgRad/dataNodeCount);
	}
	
}
