package experimentalcode.tobias._index.helpstructures;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.PriorityQueue;

public class KNNList<E> {
  PriorityQueue<DistanceObjectTuple<E>> pq;

  int k;

  public KNNList(int k) {
    pq = new PriorityQueue<DistanceObjectTuple<E>>();
    this.k = k;
  }

  public void add(E object, double distance) {
    if(pq.size() < k) {
      pq.add(new DistanceObjectTuple<E>(distance, object));
    }
    else if(pq.peek().distance > distance) {
      pq.poll();
      pq.add(new DistanceObjectTuple<E>(distance, object));
    }
  }

  /**
   * Does not delete the entries from the KNNList. Not ordered by distance.
   * 
   * @return
   */
  public ArrayList<E> getKNNs() {
    ArrayList<E> res = new ArrayList<E>(pq.size());
    Iterator<DistanceObjectTuple<E>> iter = pq.iterator();
    while(iter.hasNext()) {
      res.add(iter.next().object);
    }
    return res;
  }

  /**
   * Does not delete the entries from the KNNList. Not ordered by distance.
   * 
   * @return
   */
  /*
   * public ArrayDBIDs getIDs(){ ArrayModifiableDBIDs res = DBIDUtil.newArray();
   * Iterator<DistanceObjectTuple<E>> iter = pq.iterator();
   * while(iter.hasNext()){ res.add(iter.next().object.getID()); } return res; }
   */
  public double getKNNdist() {
    if(pq.size() < k) {
      return Double.MAX_VALUE;
    }
    return pq.peek().distance;
  }

  class DistanceObjectTuple<O> implements Comparable<DistanceObjectTuple<O>> {
    public double distance;

    public O object;

    public DistanceObjectTuple(double distance, O object) {
      this.distance = distance;
      this.object = object;
    }

    public int compareTo(DistanceObjectTuple<O> dot) {
      if(distance < dot.distance) {
        return 1;
      }
      else if(distance > dot.distance) {
        return -1;
      }
      else {
        return 0;
      }
    }
  }
}
