package experimentalcode.tobias.rknnjoin.structures;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map.Entry;

import de.lmu.ifi.dbs.elki.data.DoubleVector;
import de.lmu.ifi.dbs.elki.database.ids.DBID;
import de.lmu.ifi.dbs.elki.result.Result;

public class RkNNJoinResult<O extends DoubleVector> implements Result{
	
	public HashMap<DBID, RkNNResult> queryResults;
	
	public RkNNJoinResult(){
		queryResults = new HashMap<DBID, RkNNResult>();
	}
	public void addResult(O query, List<O> result){
		queryResults.put(query.getID(), new RkNNResult(query, result));
	}
	public void addSingleResult(O query, O result){
		if(queryResults.containsKey(query.getID())){
			queryResults.get(query.getID()).addResult(result);
		}else{
			RkNNResult r = new RkNNResult(query);
			r.addResult(result);
			queryResults.put(query.getID(), r);	
		}
	}
	public void merge(RkNNJoinResult<O> res){
		queryResults.putAll(res.queryResults);
	}

	public int getNumberOfResults(){
		int total = 0;
		for(Entry<DBID, RkNNResult> e: queryResults.entrySet()){
			total += e.getValue().rknns.size();
		}
		return total;
	}
	public class RkNNResult{
		public O query;
		public List<O> rknns;
		
		public RkNNResult(O query, List<O> rknns){
			this.query = query;
			this.rknns = rknns;
		}
		public RkNNResult(O query){
			this.query = query;
			rknns = new ArrayList<O>();
		}
		public void addResult(O result){
			rknns.add(result);
		}
	}

  @Override
  public String getLongName() {
    return "rknnjoinresult";
  }

  @Override
  public String getShortName() {
    return "rknnjoinresult";
  }
}
