package experimentalcode.tobias.rknnjoin.structures;

import java.util.List;

import de.lmu.ifi.dbs.elki.data.DoubleVector;
import de.lmu.ifi.dbs.elki.data.HyperBoundingBox;
import de.lmu.ifi.dbs.elki.distance.distancefunction.EuclideanDistanceFunction;

/**
 * Represents a group of DoubleVectors
 * @author emrich
 *
 */
public class VectorGroup{
		public List<DoubleVector> vectors;
		
		public DoubleVector center;
		public double radius;
		public HyperBoundingBox mbr;
		
		public VectorGroup(List<DoubleVector> vs, DoubleVector c, double r){
			center = c;
			vectors = vs;
			radius = r;
			setMBR();
		}
		
		public VectorGroup(List<DoubleVector> vs,DoubleVector c){
			EuclideanDistanceFunction euc = EuclideanDistanceFunction.STATIC;
			center = c;
			vectors = vs;
			for(DoubleVector v:vs){
				radius = Math.max(radius, euc.distance(v, c).getValue());
			}
			setMBR();
		}
		
		public VectorGroup(List<DoubleVector> vs){
			EuclideanDistanceFunction euc = EuclideanDistanceFunction.STATIC;
			int dim = vs.get(0).getDimensionality();
			
			double[] c = new double[dim];
			for(DoubleVector v:vs){
				for(int d = 0; d< dim;d++){
					c[d] += v.getValue(d+1);
				}
			}
			for(int d = 0; d< dim;d++){
				c[d] /= vs.size();
			}
			center = new DoubleVector(c);
			vectors = vs;
			for(DoubleVector v:vs){
				radius = Math.max(radius, euc.distance(v, center).getValue());
			}
			setMBR();
		}
		public void setMBR(){
			
			if(vectors.size() == 0){
				mbr = null;
				return;
			}
			
			double[] low, high;
			low = vectors.get(0).getValues();
			high = vectors.get(0).getValues();
			for(DoubleVector v: vectors){
				for(int d = 0; d< vectors.get(0).getDimensionality();d++){
					low[d] = Math.min(low[d], v.getValue(d+1));
					high[d] = Math.max(high[d], v.getValue(d+1));
				}
			}
			mbr = new HyperBoundingBox(low, high);
		}
		public HyperBoundingBox getMBR(){
			return mbr;
		}
}
